@@ -1,36 +1,37 @@
 # -*- coding: utf-8 -*-
-#
-# a zkCli.sh clone - though not everything is supported currently
-# It supports the basic ops:
-#
-#  python contrib/shell.py localhost:2181
-#  (CONNECTED) /> ls
-#  zookeeper
-#  (CONNECTED) /> create foo 'bar'
-#  (CONNECTED) /> get foo
-#  bar
-#  (CONNECTED) /> cd foo
-#  (CONNECTED) /foo> create ish 'barish'
-#  (CONNECTED) /foo> cd ..
-#  (CONNECTED) /> ls foo
-#  ish
-#  (CONNECTED) /> create temp- 'temp' true true
-#  (CONNECTED) /> ls
-#  zookeeper foo temp-0000000001
-#  (CONNECTED) /> rmr foo
-#  (CONNECTED) />
-#  (CONNECTED) /> tree
-#  .
-#  ├── zookeeper
-#  │   ├── config
-#  │   ├── quota
 
+"""
+zkCli.sh clone.
+It supports the basic ops plus a few handy extensions:
+ (CONNECTED) /> ls
+ zookeeper
+ (CONNECTED) /> create foo 'bar'
+ (CONNECTED) /> get foo
+ bar
+ (CONNECTED) /> cd foo
+ (CONNECTED) /foo> create ish 'barish'
+ (CONNECTED) /foo> cd ..
+ (CONNECTED) /> ls foo
+ ish
+ (CONNECTED) /> create temp- 'temp' true true
+ (CONNECTED) /> ls
+ zookeeper foo temp-0000000001
+ (CONNECTED) /> rmr foo
+ (CONNECTED) />
+ (CONNECTED) /> tree
+ .
+ ├── zookeeper
+ │   ├── config
+ │   ├── quota
+"""
 
 from __future__ import print_function
 
+from functools import wraps
 import os
 import re
 import shlex
+import signal
 import sys
 import time
 import zlib
@@ -39,11 +40,12 @@ from kazoo.exceptions import (
     BadVersionError,
     InvalidACLError,
     NoAuthError,
+    NodeExistsError,
     NoNodeError,
     NotEmptyError,
     ZookeeperError,
 )
-from kazoo.handlers.threading import TimeoutError
+from kazoo.protocol.states import KazooState
 from kazoo.security import OPEN_ACL_UNSAFE, READ_ACL_UNSAFE
 
 from .acl import ACLReader
@@ -63,51 +65,105 @@ from .watch_manager import get_watch_manager
 from .util import pretty_bytes, to_bool
 
 
+def connected(func):
+    """ check connected, fails otherwise """
+    @wraps(func)
+    def wrapper(*args, **kwargs):
+        self = args[0]
+        if not self.connected:
+            print("Not connected.", file=self.output)
+        else:
+            try:
+                return func(*args, **kwargs)
+            except NoAuthError:
+                print("Not authenticated.", file=self.output)
+    return wrapper
+
+
+def check_path_exists(func):
+    """ check paths exists """
+    @wraps(func)
+    def wrapper(*args):
+        self = args[0]
+        params = args[1]
+        path = params.path
+        params.path = self.resolve_path(path)
+        if self.client.exists(params.path):
+            return func(self, params)
+        print("Path %s doesn't exist" % (path), file=self.output)
+        return False
+    return wrapper
+
+
+def check_path_absent(func):
+    """ check path doesn't exist """
+    @wraps(func)
+    def wrapper(*args):
+        self = args[0]
+        params = args[1]
+        path = params.path
+        params.path = self.resolve_path(path)
+        if not self.client.exists(params.path):
+            return func(self, params)
+        print("Path %s already exists" % (path))
+    return wrapper
+
+
+def default_watcher(watched_event):
+    print(str(watched_event))
+
+
+# pylint: disable=R0904
 class Shell(AugumentedCmd):
-    def __init__(self, hosts=[], timeout=10, output=sys.stdout, setup_readline=True):
+    """ main class """
+    def __init__(self, hosts=None, timeout=10, output=sys.stdout, setup_readline=True, async=True):
         AugumentedCmd.__init__(self, ".kz-shell-history", setup_readline)
-        self._hosts = hosts
+        self._hosts = hosts if hosts else []
         self._connect_timeout = timeout
         self._output = output
         self._zk = None
         self._read_only = False
+        self._async = async
         self.connected = False
 
-        if len(self._hosts) > 0: self._connect(self._hosts)
-        if not self.connected: self.update_curdir("/")
+        if len(self._hosts) > 0:
+            self._connect(self._hosts)
+        if not self.connected:
+            self.update_curdir("/")
 
-    def connected(f):
-        def wrapped(self, args):
-            if not self.connected:
-                print("Not connected.", file=self._output)
-            else:
-                try:
-                    return f(self, args)
-                except NoAuthError as ex:
-                    print("Not authenticated.", file=self._output)
-        wrapped.__doc__ = f.__doc__
-        return wrapped
+    def _complete_path(self, cmd_param_text, full_cmd, *_):
+        """ completes paths """
+        pieces = shlex.split(full_cmd)
+        cmd_param = pieces[1] if len(pieces) > 1 else cmd_param_text
+        offs = len(cmd_param) - len(cmd_param_text)
+        path = cmd_param[:-1] if cmd_param.endswith("/") else cmd_param
 
-    def check_path_exists(f):
-        def wrapped(self, params):
-            path = params.path
-            params.path = self.abspath(path if path not in ["", "."] else self.curdir)
-            if self._zk.exists(params.path):
-                return f(self, params)
-            print("Path %s doesn't exist" % (path), file=self._output)
-            return False
-        wrapped.__doc__ = f.__doc__
-        return wrapped
+        if re.match(r"^\s*$", path):
+            return self._zk.get_children(self.curdir)
 
-    def check_path_absent(f):
-        def wrapped(self, params):
-            path = params.path
-            params.path = self.abspath(path if path != '' else self.curdir)
-            if not self._zk.exists(params.path):
-                return f(self, params)
-            print("Path %s already exists" % (path))
-        wrapped.__doc__ = f.__doc__
-        return wrapped
+        if self._zk.exists(path):
+            children = self._zk.get_children(self.resolve_path(path))
+            opts = ["%s/%s" % (path, znode) for znode in children]
+        elif "/" not in path:
+            znodes = self._zk.get_children(self.curdir)
+            opts = [znode for znode in znodes if znode.startswith(path)]
+        else:
+            parent = os.path.dirname(path)
+            child = os.path.basename(path)
+            matching = [znode for znode in self._zk.get_children(parent) if znode.startswith(child)]
+            opts = ["%s/%s" % (parent, znode) for znode in matching]
+
+        return [opt[offs:] for opt in opts]
+
+    @property
+    def client(self):
+        """ the connected ZK client, if any """
+        return self._zk
+
+    @property
+    def output(self):
+        """ the io output object """
+        return self._output
 
     @connected
     @ensure_params(Required("scheme"), Required("credential"))
@@ -140,8 +196,7 @@ class Shell(AugumentedCmd):
         except (NoNodeError, BadVersionError, InvalidACLError, ZookeeperError) as ex:
             print("Failed to set ACLs: %s. Error: %s" % (str(acls), str(ex)), file=self._output)
 
-    def complete_set_acls(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_set_acls = _complete_path
 
     @connected
     @interruptible
@@ -164,31 +219,29 @@ class Shell(AugumentedCmd):
         /zookeeper/config: [ACL(perms=31, acl_list=['ALL'], id=Id(scheme=u'world', id=u'anyone'))]
         /zookeeper/quota: [ACL(perms=31, acl_list=['ALL'], id=Id(scheme=u'world', id=u'anyone'))]
         """
-        def replace(l, oldv, newv):
+        def replace(plist, oldv, newv):
             try:
-                l.remove(oldv)
-                l.insert(0, newv)
+                plist.remove(oldv)
+                plist.insert(0, newv)
             except ValueError:
                 pass
 
-        for p, acls in self._zk.get_acls_recursive(params.path, params.depth, params.ephemerals):
+        for path, acls in self._zk.get_acls_recursive(params.path, params.depth, params.ephemerals):
             replace(acls, READ_ACL_UNSAFE[0], "WORLD_READ")
             replace(acls, OPEN_ACL_UNSAFE[0], "WORLD_ALL")
-            print("%s: %s" % (p, acls), file=self._output)
+            print("%s: %s" % (path, acls), file=self._output)
 
-    def complete_get_acls(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_get_acls = _complete_path
 
     @connected
     @ensure_params(Optional("path"), Optional("watch"))
     @check_path_exists
     def do_ls(self, params):
-        kwargs = {"watch": self.watcher} if to_bool(params.watch) else {}
+        kwargs = {"watch": default_watcher} if to_bool(params.watch) else {}
         znodes = self._zk.get_children(params.path, **kwargs)
         print(" ".join(znodes), file=self._output)
 
-    def complete_ls(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_ls = _complete_path
 
     @connected
     @interruptible
@@ -213,13 +266,14 @@ class Shell(AugumentedCmd):
             try:
                 repeat = int(params.debug)
                 sleep = int(params.sleep)
-            except ValueError: pass
+            except ValueError:
+                pass
             if repeat == 0:
                 while True:
                     wm.stats(params.path)
                     time.sleep(sleep)
             else:
-                for i in xrange(0, repeat):
+                for _ in range(0, repeat):
                     wm.stats(params.path)
                     time.sleep(sleep)
         else:
@@ -227,7 +281,8 @@ class Shell(AugumentedCmd):
 
     @ensure_params(Required("src"), Required("dst"),
                    BooleanOptional("recursive"), BooleanOptional("overwrite"),
-                   BooleanOptional("async"), BooleanOptional("verbose"))
+                   BooleanOptional("async"), BooleanOptional("verbose"),
+                   IntegerOptional("max_items", 0))
     def do_cp(self, params):
         """
         copy from/to local/remote or remote/remote paths.
@@ -246,7 +301,7 @@ class Shell(AugumentedCmd):
 
         examples:
         cp /some/znode /backup/copy-znode  # local
-        cp file://<path> zk://[user:passwd@]host/<path> <recursive> <overwrite> <async> <verbose>
+        cp file://<path> zk://[user:passwd@]host/<path> <recursive> <overwrite> <async> <verbose> <max_items>
         cp /some/path json://!home!user!backup.json/ true true
         """
 
@@ -256,31 +311,33 @@ class Shell(AugumentedCmd):
             zk_url = self._zk.zk_url()
 
             # if these are local paths, make them absolute paths
-            if not re.match("^\w+://", params.src):
-                params.src = "%s%s" % (zk_url, self.abspath(params.src))
+            if not re.match(r"^\w+://", params.src):
+                params.src = "%s%s" % (zk_url, self.resolve_path(params.src))
                 src_connected_zk = True
 
-            if not re.match("^\w+://", params.dst):
-                params.dst = "%s%s" % (zk_url, self.abspath(params.dst))
+            if not re.match(r"^\w+://", params.dst):
+                params.dst = "%s%s" % (zk_url, self.resolve_path(params.dst))
                 dst_connected_zk = True
 
         try:
-            src = Proxy.from_string(params.src, True)
+            src = Proxy.from_string(params.src, True, params.async, params.verbose)
             if src_connected_zk:
                 src.need_client = False
                 src.client = self._zk
 
-            dst = Proxy.from_string(params.dst, None if params.overwrite else False)
+            dst = Proxy.from_string(params.dst,
+                                    exists=None if params.overwrite else False,
+                                    async=params.async,
+                                    verbose=params.verbose)
             if dst_connected_zk:
                 dst.need_client = False
                 dst.client = self._zk
 
-            src.copy(dst, params.recursive, params.async, params.verbose)
+            src.copy(dst, params.recursive, params.max_items)
         except CopyError as ex:
             print(str(ex), file=self._output)
 
-    def complete_cp(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_cp = _complete_path
 
     @connected
     @interruptible
@@ -303,11 +360,10 @@ class Shell(AugumentedCmd):
         ├── bar
         """
         print(".", file=self._output)
-        for c, l in self._zk.tree(params.path, params.max_depth):
-            print(u"%s├── %s" % (u"│   " * l, c), file=self._output)
+        for child, level in self._zk.tree(params.path, params.max_depth):
+            print(u"%s├── %s" % (u"│   " * level, child), file=self._output)
 
-    def complete_tree(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_tree = _complete_path
 
     @connected
     @ensure_params(Optional("path"))
@@ -328,11 +384,10 @@ class Shell(AugumentedCmd):
         /fooish/xorg
         /copy/foo
         """
-        for p in self._zk.find(params.path, params.match, 0):
-            print(p, file=self._output)
+        for path in self._zk.find(params.path, params.match, 0):
+            print(path, file=self._output)
 
-    def complete_find(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_find = _complete_path
 
     @connected
     @ensure_params(Required("path"), Required("match"))
@@ -347,11 +402,10 @@ class Shell(AugumentedCmd):
         /fooish/xorg
         /copy/Foo
         """
-        for p in self._zk.find(params.path, params.match, re.IGNORECASE):
-            print(p, file=self._output)
+        for path in self._zk.find(params.path, params.match, re.IGNORECASE):
+            print(path, file=self._output)
 
-    def complete_ifind(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_ifind = _complete_path
 
     @connected
     @ensure_params(Required("path"), Required("content"), BooleanOptional("show_matches"))
@@ -366,8 +420,7 @@ class Shell(AugumentedCmd):
         """
         self.grep(params.path, params.content, 0, params.show_matches)
 
-    def complete_grep(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_grep = _complete_path
 
     @connected
     @ensure_params(Required("path"), Required("content"), BooleanOptional("show_matches"))
@@ -382,17 +435,16 @@ class Shell(AugumentedCmd):
         """
         self.grep(params.path, params.content, re.IGNORECASE, params.show_matches)
 
-    def complete_igrep(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_igrep = _complete_path
 
     def grep(self, path, content, flags, show_matches):
-        for p, matches in self._zk.grep(path, content, flags):
+        for path, matches in self._zk.grep(path, content, flags):
             if show_matches:
-                print("%s:" % (p), file=self._output)
-                for m in matches:
-                    print(m, file=self._output)
+                print("%s:" % (path), file=self._output)
+                for match in matches:
+                    print(match, file=self._output)
             else:
-                print(p, file=self._output)
+                print(path, file=self._output)
 
     @connected
     @ensure_params(Required("path"))
@@ -400,8 +452,7 @@ class Shell(AugumentedCmd):
     def do_cd(self, params):
         self.update_curdir(params.path)
 
-    def complete_cd(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_cd = _complete_path
 
     @connected
     @ensure_params(Required("path"), Optional("watch"))
@@ -421,18 +472,18 @@ class Shell(AugumentedCmd):
         set /foo 'notbar'
         WatchedEvent(type='CHANGED', state='CONNECTED', path=u'/foo')
         """
-        kwargs = {"watch": self.watcher} if to_bool(params.watch) else {}
-        value, stat = self._zk.get(params.path, **kwargs)
+        kwargs = {"watch": default_watcher} if to_bool(params.watch) else {}
+        value, _ = self._zk.get(params.path, **kwargs)
 
         # maybe it's compressed?
         try:
             value = zlib.decompress(value)
-        except (zlib.error, TypeError): pass
+        except (zlib.error, TypeError):
+            pass
 
         print(value, file=self._output)
 
-    def complete_get(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_get = _complete_path
 
     @connected
     @ensure_params(Required("path"), Optional("watch"))
@@ -451,18 +502,15 @@ class Shell(AugumentedCmd):
         rm /foo
         WatchedEvent(type='DELETED', state='CONNECTED', path=u'/foo')
         """
-        kwargs = {"watch": self.watcher} if to_bool(params.watch) else {}
-        stat = self._zk.exists(params.path, **kwargs)
+        kwargs = {"watch": default_watcher} if to_bool(params.watch) else {}
+        path = self.resolve_path(params.path)
+        stat = self._zk.exists(path, **kwargs)
         if stat:
             print(stat, file=self._output)
         else:
             print("Path %s doesn't exist" % (params.path), file=self._output)
 
-    def complete_exists(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
-
-    def watcher(self, watched_event):
-        print((str(watched_event)))
+    complete_exists = _complete_path
 
     @connected
     @ensure_params(Required("path"),
@@ -498,15 +546,17 @@ class Shell(AugumentedCmd):
         │   │   ├── path
         │   │   │   ├── here
         """
-        self._zk.create(params.path,
-                        params.value,
-                        acl=None,
-                        ephemeral=params.ephemeral,
-                        sequence=params.sequence,
-                        makepath=params.recursive)
+        try:
+            self._zk.create(params.path,
+                            params.value,
+                            acl=None,
+                            ephemeral=params.ephemeral,
+                            sequence=params.sequence,
+                            makepath=params.recursive)
+        except NodeExistsError:
+            print("Path %s exists" % (params.path), file=self._output)
 
-    def complete_create(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_create = _complete_path
 
     @connected
     @ensure_params(Required("path"), Required("value"))
@@ -520,8 +570,7 @@ class Shell(AugumentedCmd):
         """
         self._zk.set(params.path, params.value)
 
-    def complete_set(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_set = _complete_path
 
     @connected
     @ensure_params(Required("path"))
@@ -532,8 +581,7 @@ class Shell(AugumentedCmd):
         except NotEmptyError:
             print("%s is not empty." % (params.path))
 
-    def complete_rm(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_rm = _complete_path
 
     @connected
     @ensure_params()
@@ -554,10 +602,10 @@ xid=%d
 last_zxid=%d
 timeout=%d
 server=%s""" % (self._zk.state,
-                self._zk._connection._xid,
+                self._zk.xid,
                 self._zk.last_zxid,
-                self._zk._session_timeout,
-                self._zk._connection._socket.getpeername()), file=self._output)
+                self._zk.session_timeout,
+                self._zk.server), file=self._output)
 
     @ensure_params(Optional("host"))
     def do_mntr(self, params):
@@ -595,8 +643,7 @@ server=%s""" % (self._zk.state,
         """
         self._zk.delete(params.path, recursive=True)
 
-    def complete_rmr(self, cmd_param_text, full_cmd, start_idx, end_idx):
-        return self._complete_path(cmd_param_text, full_cmd)
+    complete_rmr = _complete_path
 
     @connected
     @ensure_params(Required("path"))
@@ -646,37 +693,31 @@ server=%s""" % (self._zk.state,
     def _connect(self, hosts):
         self._disconnect()
         self._zk = AugumentedClient(",".join(hosts), read_only=self._read_only)
+        if self._async:
+            self._connect_async()
+        else:
+            self._connect_sync()
+
+    def _connect_async(self):
+        def listener(state):
+            if state == KazooState.CONNECTED:
+                self.connected = True
+                self.update_curdir("/")
+                # hack to restart sys.stdin.readline()
+                print("", file=self.output)
+                os.kill(os.getpid(), signal.SIGUSR2)
+        self._zk.add_listener(listener)
+        self._zk.start_async()#timeout=self._connect_timeout)
+        self.update_curdir("/")
+
+    def _connect_sync(self):
         try:
             self._zk.start(timeout=self._connect_timeout)
             self.connected = True
-        except TimeoutError as ex:
+        except self._zk.handler.timeout_exception as ex:
             print("Failed to connect: %s" % (ex), file=self._output)
-
         self.update_curdir("/")
 
     @property
     def state(self):
         return "(%s) " % (self._zk.state if self._zk else "DISCONNECTED")
-
-    def _complete_path(self, cmd_param_text, full_cmd):
-        pieces = shlex.split(full_cmd)
-        cmd_param = pieces[1] if len(pieces) > 1 else cmd_param_text
-        offs = len(cmd_param) - len(cmd_param_text)
-        path = cmd_param[:-1] if cmd_param.endswith("/") else cmd_param
-
-        if re.match("^\s*$", path):
-            return self._zk.get_children(self.curdir)
-
-        if self._zk.exists(path):
-            children = self._zk.get_children(self.abspath(path))
-            opts = list(map(lambda z: "%s/%s" % (path, z), children))
-        elif "/" not in path:
-            znodes = self._zk.get_children(self.curdir)
-            opts = list(filter(lambda z: z.startswith(path), znodes))
-        else:
-            parent = os.path.dirname(path)
-            child = os.path.basename(path)
-            matching = list(filter(lambda z: z.startswith(child), self._zk.get_children(parent)))
-            opts = list(map(lambda z: "%s/%s" % (parent, z), matching))
-
-        return list(map(lambda x: x[offs:], opts))
